/* 
 * gs_filefunc.c
 *
 * Common file functions
 * 
 * by immi ( immi at gmx . org )
 * by eis_os (cim bytetransfer.de)
 *
 * last update: 27.05.2012
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <errno.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "gs.h"
#include "gs_filefunc.h"

void print(unsigned int indent_level, const char* fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    printf("%*s", indent_level*2, "");
    vprintf(fmt, ap);
    fflush(stdout);
    va_end(ap);
}


void die(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    fflush(stdout);
    fflush(stderr);
    fprintf(stderr, "*error*:");
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exit(-1);
}


void die_parser(struct parser* parser, const char* p, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    fflush(stdout);
    fflush(stderr);
    fprintf(stderr, "*error*:");
    vfprintf(stderr, fmt, ap);
    
    
    if (p != NULL) {
        int pos = p-parser->line;
        fprintf(stderr, "on line %u,%u:\n%s\n", parser->nline, pos, parser->line);
        fprintf(stderr, "%*s\n", pos, "^");
    } else {
        fprintf(stderr, "on line %u: %s\n", parser->nline, parser->line);
    }
    if (parser->errinfo != NULL) {
        fprintf(stderr, "%s\n", parser->errinfo);
    }
    
    va_end(ap);
    exit(-1);
}


// returns 1 if true/zero if false
int file_exists(const char *file)
{
    struct stat st;
    if(stat(file, &st) == 0
       && S_ISREG(st.st_mode)) return 1;
    return 0;
}

char* make_path(const char *path, const char* file, const char *suffix)
{
    char *full;
    unsigned int len = strlen(file) + 1;

    if(path)
	len += strlen(path) + 1;
    if(suffix)
	len += strlen(suffix);
    full = malloc(len);

    if(path)
	sprintf(full, "%s%c%s%s", path, PATH_DELIMITER, file,
		suffix ? suffix : "");
    else
	sprintf(full, "%s%s", file, suffix ? suffix : "");
    return full;
}

char* filename_noext(const char *file)
{
    char *basename, *dotpos;
    int len;
    dotpos = strrchr(file, '.');
    if (dotpos != NULL) {
        len = dotpos-file;
        basename = malloc(len+1);
        basename[len] = 0;
        strncpy(basename, file, len);
    } else {
        basename = strdup(file);
    }   
    return basename;
}


int gs_read(int fd, void* dst, unsigned int len, unsigned int *remaining)
{
    ssize_t nb;

    if(*remaining < len)
	die("trying to read beyond chunk end(req: %u/avail:%u)\n",
	    len, *remaining);

    nb = read(fd, dst, len);
    if(nb < 0)
	die("read(): %s\n", strerror(errno));
    else if(nb == 0)
	die("unexpected EOF\n");

    *remaining -= len; //nb;
    return 0;
}

int gs_read_chunk(int fd, struct gs_chunk *ch, unsigned int *remaining)
{
    gs_read(fd, ch, sizeof(*ch), remaining);

    if(ch->size > *remaining)
	die("chunk: exceeds size of parents data block\n");

    *remaining -= ch->size;
    return 0;
}


struct chunkfilepos {
    uint32_t chunktype;
    int fd;
    int pos;
};

struct chunkfilepos_f {
    uint32_t chunktype;
    FILE* file;
    int pos;
};

void* write_chunk_start(int fd, uint32_t chunktype)
{
    struct gs_chunk ch;
    struct chunkfilepos* state;
    state = malloc(sizeof(struct chunkfilepos));
    if (!state) {
        die("chunkfilepos malloc failed");
    }
    state->chunktype = chunktype;
    state->fd = fd;
    state->pos = tell(fd);
    
    // Write a dummy chunk
    ch.type = 0;
    ch.size = 0;
    write(fd, &ch, sizeof(ch));
    return state;
}

unsigned int write_chunk_end(void* s)
{
    struct chunkfilepos* state = s;
    struct gs_chunk ch;
    int curpos;
    

    curpos = tell(state->fd);
    
    lseek(state->fd, state->pos, SEEK_SET);
    
    ch.type = state->chunktype;
    ch.size = (curpos-state->pos-sizeof(ch));
    write(state->fd, &ch, sizeof(ch));
    
    lseek(state->fd, curpos, SEEK_SET);
    free(state);
    return ch.size;
}


void* fwrite_chunk_start(FILE* f, uint32_t chunktype)
{
    struct gs_chunk ch;
    struct chunkfilepos_f* state;
    state = malloc(sizeof(struct chunkfilepos));
    if (!state) {
        die("chunkfilepos malloc failed");
    }
    state->chunktype = chunktype;
    state->file = f;
    state->pos = ftell(f);
    
    // Write a dummy chunk
    ch.type = 0;
    ch.size = 0;
    fwrite(&ch, sizeof(ch), 1, f);
    return state;
}

unsigned int fwrite_chunk_end(void* s)
{
    struct chunkfilepos_f* state = s;
    struct gs_chunk ch;
    int curpos;
    

    curpos = ftell(state->file);
    
    fseek(state->file, state->pos, SEEK_SET);
    
    ch.type = state->chunktype;
    ch.size = (curpos-state->pos-sizeof(ch));
    fwrite(&ch, sizeof(ch), 1, state->file);
    
    fseek(state->file, curpos, SEEK_SET);
    free(state);
    return ch.size;
}



// Parser



float read_float(const char *p, char **end, struct parser* parser)
{
    float res;

    errno = 0;
    res = strtof(p, end);
    if(errno)
	die_parser(parser,p,"strtof(): %s\n", strerror(errno));
    if(*end == p)
	die_parser(parser,p,"expected float\n");
    return res;
}

unsigned int read_uint(char *p, char **end, struct parser* parser)
{
    unsigned int res;
    errno = 0;
    res = strtoul(p, end, 10);
    if(errno)
	die_parser(parser,p,"strtoul(): %s\n", strerror(errno));
    if(p == *end)
	die_parser(parser,p,"parser: expected integer\n");
    return res;
}

void parser_fill_buffer(struct parser* parser)
{
    ssize_t nb;

    if(parser->bytes_left)
	memmove(parser->buf, parser->cur, parser->bytes_left);

    nb = read(parser->fd, parser->buf + parser->bytes_left,
	      BUFSIZE - 1 - parser->bytes_left);
    if(nb < 0)
	die("read(): %s\n", strerror(errno));

    parser->bytes_left = nb + parser->bytes_left;
    parser->buf[parser->bytes_left] = '\0';
    parser->cur = parser->buf;
}

void parser_read_line(struct parser *parser)
{
    char *p;

    p = strchr(parser->cur, '\n');
    if(!p) {
	parser_fill_buffer(parser);
	p = strchr(parser->cur, '\n');
	if(!p) {
	    parser->line = NULL; // EOF
	    return;
	}
    }
    parser->line = parser->cur;

    /* zero out EOL so we have a nicely terminated string */
    if(p > parser->cur) // handle \r\n (unless \n is the first char in line)
        if( *(p-1) == '\r') *(p-1) = '\0';
    *p = 0;

    parser->bytes_left -= (p+1) - parser->cur;
    parser->cur = p + 1;
    parser->nline++;
    parser->errinfo = NULL;
}

struct parser* init_parser(const char* filename)
{
    struct parser *parser;
    int fd;

    fd = open(filename, OPEN_FLAGS);
    if(fd == -1)
	die("open(%s) failed: %s\n", filename, strerror(errno));

    parser = malloc(sizeof(*parser));
    memset(parser, 0, sizeof(*parser));
    parser->fd = fd;

    parser_fill_buffer(parser);
    parser_read_line(parser); // read first line

    printf("reading: '%s' ...\n", filename);
    return parser;
}


